from io import BytesIO

from airflow.providers.amazon.aws.hooks.s3 import S3Hook

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol


class S3Filesystem(FilesystemProtocol):
    def __init__(self, hook: S3Hook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        bucket_name, key_name = _get_bucket_and_key_name(path)
        obj = self.hook.get_key(key_name, bucket_name)
        return obj.get()["Body"]

    def write(self, data: str | bytes | BytesIO, path: str):
        bucket_name, key_name = _get_bucket_and_key_name(path)
        if isinstance(data, str):
            self.conn.load_string(data, key_name, bucket_name)
        elif isinstance(data, bytes):
            self.conn.load_bytes(data, key_name, bucket_name)
        else:
            self.conn.load_file_obj(data, key_name, bucket_name)

    def delete_prefix(self, prefix: str):
        bucket_name, key_prefix = _get_bucket_and_key_name(prefix)
        self.conn.get_bucket(bucket_name).objects.filter(Prefix=key_prefix).delete()


def _get_bucket_and_key_name(path: str) -> tuple[str, str]:
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])



from io import BytesIO
from typing import Protocol


class FilesystemProtocol(Protocol):
    def read(self, path: str) -> bytes:
        ...

    def write(self, data: str | bytes | BytesIO, path: str):
        ...

    def delete_prefix(self, prefix: str):
        ...



from airflow.hooks.filesystem import FSHook
from airflow.models import Connection
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.microsoft.azure.hooks.wasb import WasbHook
from airflow.providers.sftp.hooks.sftp import SFTPHook

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol
from airflow_tools.filesystems.impl.blob_storage_filesystem import BlobStorageFilesystem
from airflow_tools.filesystems.impl.local_filesystem import LocalFilesystem
from airflow_tools.filesystems.impl.s3_filesystem import S3Filesystem
from airflow_tools.filesystems.impl.sftp_filesystem import SFTPFilesystem


class FilesystemFactory:
    @staticmethod
    def get_data_lake_filesystem(connection: Connection) -> FilesystemProtocol:
        if connection.conn_type == "wasb":
            hook = WasbHook(wasb_conn_id=connection.conn_id)
            return BlobStorageFilesystem(hook)
        elif connection.conn_type == "aws":
            hook = S3Hook(aws_conn_id=connection.conn_id)
            return S3Filesystem(hook)
        elif connection.conn_type == "sftp":
            hook = SFTPHook(ssh_conn_id=connection.conn_id)
            return SFTPFilesystem(hook)
        elif connection.conn_type == "fs":
            hook = FSHook(fs_conn_id=connection.conn_id)
            return LocalFilesystem(hook)
        else:
            raise NotImplementedError(
                f"Data Lake type {connection.conn_type} is not supported"
            )



from typing import Optional, Protocol

from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator

from airflow_tools.filesystems.filesystem_factory import FilesystemFactory


class Transformation(Protocol):
    def __call__(self, data: bytes, filename: str, context: dict) -> bytes:
        ...


class FilesystemToFilesystem(BaseOperator):
    """
    Copies a file from a filesystem to another filesystem.
    """

    template_fields = ('source_file_path', 'destination_path')

    def __init__(
        self,
        source_fs_conn_id: str,
        destination_fs_conn_id: str,
        source_file_path: str,
        destination_path: str,
        data_transformation: Optional[Transformation] = None,
        *args,
        **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.source_fs_conn_id = source_fs_conn_id
        self.destination_fs_conn_id = destination_fs_conn_id
        self.source_file_path = source_file_path
        self.destination_path = destination_path
        self.data_transformation = data_transformation

    def execute(self, context):
        source_fs_hook = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.source_fs_conn_id),
        )
        destination_fs_hook = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.destination_fs_conn_id),
        )

        file_name = self.source_file_path.split('/')[-1]
        data = source_fs_hook.read(self.source_file_path)
        if self.data_transformation:
            data = self.data_transformation(data, file_name, context)
        full_destination_path = (
            self.destination_path + file_name
            if self.destination_path.endswith('/')
            else self.destination_path
        )
        destination_fs_hook.write(data, full_destination_path)



import logging
from io import BytesIO
from typing import Union

from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.microsoft.azure.hooks.wasb import WasbHook

DataLakeConnection = Union[WasbHook, S3Hook, AwsGenericHook]
logger = logging.getLogger(__file__)


class DataLakeFacade:
    """Provides a consistent interface over different Data Lakes (S3/Blob Storage etc.)"""

    def __init__(self, conn: Union[WasbHook, S3Hook, AwsGenericHook]):
        self.conn: S3Hook | WasbHook = (
            S3Hook(conn.aws_conn_id)
            if isinstance(conn, AwsGenericHook) and not isinstance(conn, S3Hook)
            else conn
        )

    def write(self, data: str | bytes | BytesIO, path: str):
        match self.conn.conn_type:
            case "wasb":
                assert isinstance(self.conn, WasbHook)
                container_name, blob_name = _get_container_and_blob_name(path)
                if isinstance(data, str):
                    self.conn.load_string(data, container_name, blob_name)
                    return

                if isinstance(data, BytesIO):
                    data = data.getvalue()
                logger.info(
                    f'Writing to wasb container "{container_name}" and blob "{blob_name}"'
                )
                self.conn.upload(
                    container_name=container_name, blob_name=blob_name, data=data
                )
            case "aws":
                assert isinstance(self.conn, S3Hook)
                bucket_name, key_name = _get_bucket_and_key_name(path)
                if isinstance(data, str):
                    self.conn.load_string(data, key_name, bucket_name)
                elif isinstance(data, bytes):
                    self.conn.load_bytes(data, key_name, bucket_name)
                else:
                    self.conn.load_file_obj(data, key_name, bucket_name)
            case _:
                raise NotImplementedError(
                    f"Data Lake type {self.conn.conn_type} does not support write"
                )

    def delete_prefix(self, prefix: str):
        match self.conn.conn_type:
            case "wasb":
                assert isinstance(self.conn, WasbHook)
                container_name, blob_prefix = _get_container_and_blob_name(prefix)
                self.conn.delete_file(
                    container_name, blob_prefix, is_prefix=True, ignore_if_missing=True
                )
            case "aws":
                assert isinstance(self.conn, S3Hook)
                bucket_name, key_prefix = _get_bucket_and_key_name(prefix)
                self.conn.get_bucket(bucket_name).objects.filter(
                    Prefix=key_prefix
                ).delete()
            case _:
                raise NotImplementedError(
                    f"Data Lake type {self.conn.conn_type} does not support delete_prefix"
                )


def _get_container_and_blob_name(path: str) -> tuple[str, str]:
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])


def _get_bucket_and_key_name(path: str) -> tuple[str, str]:
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])



import shutil
from io import BytesIO
from pathlib import Path

from airflow.hooks.filesystem import FSHook

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol


class LocalFilesystem(FilesystemProtocol):
    def __init__(self, hook: FSHook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        (Path(self.hook.get_path()) / path).read_bytes()

    def write(self, data: str | bytes | BytesIO, path: str):
        if isinstance(data, str):
            data = data.encode()
        elif isinstance(data, BytesIO):
            data = data.getvalue()
        (Path(self.hook.get_path()) / path).write_bytes(data)

    def delete_prefix(self, prefix: str):
        path_to_delete = Path(self.hook.get_path()) / prefix
        shutil.rmtree(str(path_to_delete))



from typing import TYPE_CHECKING

from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator

from airflow_tools.data_lake_facade import DataLakeFacade

if TYPE_CHECKING:
    from airflow.utils.context import Context


class DataLakeDeleteOperator(BaseOperator):
    template_fields = ('data_lake_path',)

    def __init__(self, data_lake_conn_id: str, data_lake_path: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_lake_conn_id = data_lake_conn_id
        self.data_lake_path = data_lake_path

    def execute(self, context: 'Context'):
        data_lake_conn = BaseHook.get_connection(self.data_lake_conn_id)
        data_lake_facade = DataLakeFacade(
            conn=data_lake_conn.get_hook(),
        )
        data_lake_facade.delete_prefix(self.data_lake_path)



import logging
from io import BytesIO

from airflow.providers.microsoft.azure.hooks.wasb import WasbHook

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol

logger = logging.getLogger(__file__)


class BlobStorageFilesystem(FilesystemProtocol):
    def __init__(self, hook: WasbHook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        container_name, blob_name = _get_container_and_blob_name(path)
        stream = self.hook.download(container_name=container_name, blob_name=blob_name)
        return stream.readall().encode()

    def write(self, data: str | bytes | BytesIO, path: str):
        container_name, blob_name = _get_container_and_blob_name(path)
        if isinstance(data, str):
            self.conn.load_string(data, container_name, blob_name)
            return

        if isinstance(data, BytesIO):
            data = data.getvalue()
        logger.info(
            f'Writing to wasb container "{container_name}" and blob "{blob_name}"'
        )
        self.conn.upload(container_name=container_name, blob_name=blob_name, data=data)

    def delete_prefix(self, prefix: str):
        container_name, blob_prefix = _get_container_and_blob_name(prefix)
        self.conn.delete_file(
            container_name, blob_prefix, is_prefix=True, ignore_if_missing=True
        )


def _get_container_and_blob_name(path: str) -> tuple[str, str]:
    parts = path.split("/")
    return parts[0], "/".join(parts[1:])
